#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "BinarySocket.h"

BinarySocket::BinarySocket() : m_socket_fd(-1),
                               m_is_connected(false),
                               m_recv(false),
                               m_connect(false)
{

}

BinarySocket::~BinarySocket()
{
    StopReceive();
    Finalize();
}

BinarySocketRet BinarySocket::Initialize()
{
    if (Pri_CreateClientSocket() != BINARY_SOCKET_RET_OK)
    {
        printf("Pri_CreateClientSocket() failed\n");
        return BINARY_SOCKET_RET_NG;
    }

    m_connect = true;
    connectTask = std::thread(&BinarySocket::Pri_ConnectSocket, this);
    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::Finalize()
{
    if (connectTask.joinable())
    {
        m_connect = false;
        pthread_cancel(connectTask.native_handle());
        connectTask.join();
    }

    if (close(m_socket_fd) < 0)
    {
        return BINARY_SOCKET_RET_NG;
    }

    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::RestartSocket()
{
    if (Pri_CloseClientSocket() != BINARY_SOCKET_RET_OK)
    {
        printf("Failed to close client socket\n");
        return BINARY_SOCKET_RET_NG;
    }

    if (Pri_CreateClientSocket() != BINARY_SOCKET_RET_OK)
    {
        printf("Failed to open client socket\n");
        return BINARY_SOCKET_RET_NG;
    }

    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::StartReceive()
{
    m_recv = true;
    recvTask = std::thread(&BinarySocket::Pri_TaskThread, this);

    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::StopReceive()
{
    if (recvTask.joinable())
    {
        m_recv = false;
        pthread_cancel(recvTask.native_handle());
        recvTask.join();
    }

    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::SendReplyMessage(std::string i_reply_msg)
{
    printf("SendReplyMessage HERE\n");
    if (send(m_socket_fd, i_reply_msg.c_str(), strlen(i_reply_msg.c_str()), MSG_NOSIGNAL) < 0)
    {
        printf("Send failed\n");
        return BINARY_SOCKET_RET_NG;
    }

    return BINARY_SOCKET_RET_OK;
}

void BinarySocket::Reconnect()
{
    m_is_connected = false;
}

void BinarySocket::RegisterCallbackFunc(std::function<void(std::string)> i_function)
{
    m_callback_func = i_function;
}

BinarySocketRet BinarySocket::Pri_CreateClientSocket()
{
    m_socket_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (m_socket_fd < 0)
    {
        printf("Socket creation error\n");
        return BINARY_SOCKET_RET_NG;
    }

    return BINARY_SOCKET_RET_OK;
}

BinarySocketRet BinarySocket::Pri_CloseClientSocket()
{
    if (m_socket_fd >= 0)
    {
        if (close(m_socket_fd) < 0)
        {
            return BINARY_SOCKET_RET_NG;
        }
    }

    return BINARY_SOCKET_RET_OK;
}

void BinarySocket::Pri_ConnectSocket()
{
    struct sockaddr_in serv_addr;
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(8080);
    serv_addr.sin_addr.s_addr = inet_addr("127.0.0.1");

    while (m_connect)
    {
        if (!m_is_connected)
        {
            if (connect(m_socket_fd, (struct sockaddr*)&serv_addr, sizeof(struct sockaddr)) < 0)
            {
                printf("Connection failed, retry...\n");
            }
            else
            {
                printf("Connected to host!\n");
                m_is_connected = true;
            }
        }
        usleep(50*1000);
    }
}

void BinarySocket::Pri_TaskThread()
{
    while (m_recv)
    {
        char buffer[1024] = {0};
        int recv_len = 0;
        recv_len = recv(m_socket_fd, &buffer, sizeof(buffer) - 1, 0);
        if (recv_len > 0)
        {
            std::string recv_command(buffer);
            if (m_callback_func)
            {
                m_callback_func(recv_command);
            }
            else
            {
                printf("Callback function not registered\n");
            }
        }
    }
}